/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * A server to accept incoming TCP socket tunnel connections.
 *
 * $Id: TCPSockTunServer.cc,v 1.1 2008-05-15 14:29:52 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include <errno.h>
#include "TCPSockTunServer.hh"
#include "Debug.hh"
#include "Callback.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "Time.hh"
#include "TCPSock.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
TCPSockTunServer::TCPSockTunServer(const SockTunCallback* userCallback) throw(TunnelError)
: SockTunServer(userCallback), m_thread(0), m_sleepCond(), m_shutdownMtx(), m_shutdownCond(),
  m_shutDown(false), m_threadMgr()
{
	DPRINTV("TCPSockTunServer::TCPSockTunServer()");

	// Create TCP socket
	m_prot = Enums::TCP;
	m_sock = new TCPSock();

	// Bind to ephemeral port on the loopback interface
	m_sock->bind(m_listenIP, m_listenPort);

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "TCPSockTunServer bound to " << m_listenIP << ":" << m_listenPort;
	DPRINTV(dbg.str());
	#endif

	// Start listening
	m_sock->listen();

	// Launch server to process TCP connections
	ZeroArgClassCallback<TCPSockTunServer> cb(&TCPSockTunServer::startServer, this);
	m_thread = m_threadMgr.newThread(cb);
}

TCPSockTunServer::~TCPSockTunServer()
{
	DPRINTV("TCPSockTunServer::~TCPSockTunServer()");

	// Take mutexes
	MutexVar mv(m_mtx);
	MutexVar smv(m_shutdownMtx);

	// Signal we are no longer running
	m_shutDown = true;

	// Wakeup server thread & wait for it to finish
	if(m_thread) {
		m_sleepCond.signal();
		mv.unlock();
		m_shutdownCond.wait(smv);
		mv.lock();
		m_thread = 0;
	}
}


/*********************/
/* Private Interface */
/*********************/


// Wait for arriving connections - Thread entry point
void TCPSockTunServer::startServer(void)
{
	DPRINTI("TCPSockTunServer::startServer() - Waiting for connection");

	MutexVar mv(m_mtx);

	while(!m_shutDown) {
		// Poll socket until we think it's worth trying to accept()
		try {
			bool pollError, pollHup;
			if(m_sock->pollRead(0, pollError, pollHup)) {
				// Accept pending connection
				Sock* connectedSock = m_sock->accept();

				// Get underlying connected socket
				int sockfd = connectedSock->release();
				delete connectedSock;

				// Pass connected socket back to user
				callbackToUser(sockfd);
			}
			else {
				// Poll timeout, wait a while before retrying
				Time pollTime(0, POLL_TIMEOUT);
				m_sleepCond.timedwait(mv, pollTime.getTimeVal());
			}
		}
		catch(SockError& e) {
			if(e.m_errno == EINTR) {
				// Interrupted system call
				continue;
			}
			else {
				// General error, just carry on
				continue;
			}
		}
	}

	// Signal we have finished
	mv.unlock();
	MutexVar smv(m_shutdownMtx);
	m_shutdownCond.signal();
	smv.unlock();

	DPRINTI("TCPSockTunServer closing down");
}


}	// namespace

