#include <signal.h>
#include <poll.h>
#include <unistd.h>
#include <cstdlib>
#include <iostream>
#include <vector>
#include <algorithm>
#include "HTTPTunneler.hh"
#include "Mutex.hh"
#include "MutexVar.hh"
#include "ConditionVar.hh"
#include "Time.hh"

using namespace HTTPTunnel;

// List of connected client socket descriptors
std::vector<int> ClientSocks;
Mutex SocksMtx;
ConditionVar SocksCond;


// Read and write between stdin and HTTP client sockets
void ReadWrite(void)
{
	bool keepRunning = true;

	while(keepRunning) {
		MutexVar mv(SocksMtx);

		// Work out how many descriptors we're dealing with
		unsigned int nofds = 1 + ClientSocks.size();

		// Setup poll struct for STDIN and the connected client sockets
		struct pollfd pollStruct[nofds];
		pollStruct[0].fd = STDIN_FILENO;
		pollStruct[0].events = POLLIN;
		pollStruct[0].revents = 0;
		for(unsigned int i = 0; i < ClientSocks.size(); ++i) {
			pollStruct[i + 1].fd = ClientSocks[i];
			pollStruct[i + 1].events = POLLIN;
			pollStruct[i + 1].revents = 0;
		}

		// Do the poll
		int readyfds = poll(pollStruct, nofds, 0);
		if(readyfds > 0) {
			// Check events
			for(unsigned int i = 0; i < nofds; ++i) {
				short revents = pollStruct[i].revents;
				if((revents & POLLERR) || (revents & POLLHUP) || (revents & POLLNVAL)) {
					if(revents & POLLERR)
						std::cerr << "Got ERR on fd ";
					else if(revents & POLLHUP)
						std::cerr << "Got HUP on fd ";
					else
						std::cerr << "Got NVAL on fd ";
					std::cerr << i << " (" << pollStruct[i].fd << ")" << std::endl;

					keepRunning = false;
					break;
				}
				else if(revents & POLLIN) {
					// Got data to read
					int fd = pollStruct[i].fd;
					char buff[1024];
					ssize_t noread = read(fd, buff, sizeof(buff));

					// Check for EOF
					if(noread == 0) {
						// Stdin closed?
						if(fd == STDIN_FILENO) {
							std::cerr << "Got EOF from STDIN - shutting down" << std::endl;

							// Close down everything
							while(ClientSocks.size() > 0) {
								close(ClientSocks[0]);
								ClientSocks.erase(ClientSocks.begin());
							}
							keepRunning = false;
						}
						else {
							std::cerr << "Got EOF from connected client ";
							std::cerr << i << " (" << pollStruct[i].fd << ")" << std::endl;

							// Tidy up closed client connection
							close(fd);
							std::vector<int>::iterator sit;
							sit = std::find(ClientSocks.begin(), ClientSocks.end(), fd);
							ClientSocks.erase(sit);
						}
					}
					else {
						// Data read from stdin?
						if(fd == STDIN_FILENO) {
							// Write data from stdin to all connected clients
							std::vector<int>::iterator sit;
							for(sit = ClientSocks.begin(); sit != ClientSocks.end(); ++sit)
								write(*sit, buff, noread);
						}
						else {
							// Write data from socket to stdout
							write(STDOUT_FILENO, buff, noread);
						}
					}
				}
			}
		}
		else if(readyfds < 0) {
			// Poll error
			std::cerr << "Error while polling" << std::endl;
			keepRunning = false;
		}
		else {
			// Poll timeout - wait a while before trying again
			Time pollTime(0,10);
			SocksCond.timedwait(mv, pollTime.getTimeVal());
		}
	}
}


// Callback to notify us of a new connected client
void NewClientConnection(int sock)
{
	std::cerr << "Got new connection from client - socket fd " << sock << std::endl;

	// Add client socket to list of connected clients
	MutexVar mv(SocksMtx);
	ClientSocks.push_back(sock);

	// Signal I/O thread it has a new socket to process
	SocksCond.signal();
}


// Main
int main(int argc, char** argv)
{
	try {
		// Configure global settings
		HTTPTunneler* tunneler = HTTPTunneler::Get();
		//GlobalConfigBlob globalBlob(VERBOSE);
		GlobalConfigBlob globalBlob(QUIET);
		tunneler->ConfigureGlobal(globalBlob);

		// Start the tunneling library
		tunneler->Start();
		std::cerr << "Tunneling library started" << std::endl;

		// Configure server tunnel socket
		SockTunServerConfigBlob svrBlob("localhost", 8080, false,
			new GlobalSockTunCallback(&NewClientConnection));
		TunnelerID svrID = tunneler->LaunchServerSocket(svrBlob);
		std::cerr << "Server tunneler ID: " << svrID << std::endl;

		// Do I/O
		ReadWrite();

		// Wait for tunneler to tidy up and become inactive
		while(tunneler->IsTunnelerActive(svrID))
			sleep(1);

		// Remove configured tunnels
		tunneler->RemoveTunneler(svrID);
		std::cerr << "Tunnels removed" << std::endl;

		// Stop the tunneling library
		tunneler->Stop();
		std::cerr << "Tunneling library stopped" << std::endl;
	}
	catch(TunnelError& e) {
		std::cerr << "Error running tunneler: " << e.what() << std::endl;
	}

	// Tidy up and exit
	try {
		HTTPTunneler::Release();
	}
	catch(TunnelError& e) {
		std::cerr << "Error releasing tunneler: " << e.what() << std::endl;
	}

	exit(0);
}


